# -*- coding: utf-8 -*-
"""request_token decorators."""
import functools
import logging

from django.http import HttpResponseForbidden

from jwt.exceptions import InvalidTokenError

from request_token.exceptions import ScopeError, TokenNotFoundError

## For rendering  a custom 403-page
from django.template import loader
from request_token.settings import FOUR03_TEMPLATE

logger = logging.getLogger(__name__)


def respond_to_error(session_key, error):
    """Log request error and generate 403 response."""
    logger.warning(
        "JWT token error in session '%s': %s",
        session_key, error
    )
    if FOUR03_TEMPLATE:
        response = HttpResponseForbidden(
            loader.render_to_string(FOUR03_TEMPLATE, context = {'token_error':'Invalid URL token: %s'%session_key})
        )
    else:
        response = HttpResponseForbidden("Invalid URL token (code: %s)" % session_key)
    response.error = error
    return response


def use_request_token(view_func=None, scope=None, required=False):
    """Decorator used to indicate that a view function supports RequestTokens.

    This decorator is used in conjunction with RequestTokens and the
    RequestTokenMiddleware. By the time that a request has passed through the
    middleware and reached the view function, if a RequestToken exists it will
    have been decoded, and the payload added to the request as `token_payload`.

    This decorator is then used to map the `sub` JWT token claim to the function
    using the 'scope' kwarg - the scope must be provided, and must match the
    scope of the request token.

    Once we have verified the scope match, we then fetch the RequestToken from
    the DB, and validate it against the max use count, and the request user.

    If the RequestToken specifies a User, then the incoming request must be
    either unauthenticated, or authenticated by the same user. If it is
    unauthenticated, then this decorator will swap in the user as the request.user
    so that the view function has the user it expects.


    If the 'required' kwarg is True, then the view function expects a valid token
    in all cases - and the decorator will raise TokenNotFoundError if one does
    not exist.

    Errors are trapped and returned as HttpResponseForbidden responses, with
    the original error as the `response.error` property.

    For more details on decorators with optional args, see:
    https://blogs.it.ox.ac.uk/inapickle/2012/01/05/python-decorators-with-optional-arguments/

    """
    assert scope not in ('', None), "Decorator scope cannot be empty."

    if view_func is None:
        return functools.partial(use_request_token, scope=scope, required=required)

    @functools.wraps(view_func)
    def inner(request, *args, **kwargs):

        token = getattr(request, 'token', None)
        try:
            if token is None:
                if required is True:
                    raise TokenNotFoundError()
                else:
                    return view_func(request, *args, **kwargs)
            else:
                if token.scope == scope:
                    response = view_func(request, *args, **kwargs)
                    token.log(request, response)
                    return response
                else:
                    raise ScopeError(
                        "RequestToken scope mismatch: '%s' != '%s'" %
                        (token.scope, scope)
                    )

        except InvalidTokenError as ex:
            # this will log the exception and return a 403
            return respond_to_error(request.session.session_key, ex)

    return inner



# -*- coding: utf-8 -*-
"""request_token admin models."""
import datetime
import json

from django.contrib import admin
from django.utils.safestring import mark_safe
from django.utils.timezone import now as tz_now

from request_token.models import RequestToken


def pretty_print(data):
    """Convert dict into formatted HTML."""
    if data is None:
        return None
    pretty = json.dumps(
        data,
        sort_keys=True,
        indent=4,
        separators=(',', ': ')
    )
    return mark_safe("<code>%s</code>" % pretty.replace(" ", "&nbsp;"))


class RequestTokenAdmin(admin.ModelAdmin):

    """Admin model for RequestToken objects."""

    list_display = (
        'user',
        'scope',
        'not_before_time',
        'expiration_time',
        'max_uses',
        'used_to_date',
        'issued_at',
        'is_valid'
    )
    readonly_fields = ('pretty_payload', 'jwt', 'issued_at')
    search_fields = ('user__first_name', 'user__username')
    raw_id_fields = ('user',)

    def pretty_payload(self, obj):
        return pretty_print(obj.claims)
    pretty_payload.short_description = "Payload"

    def jwt(self, obj):
        try:
            return obj.jwt()
        except:
            return None
    jwt.short_description = "JWT"

    def is_valid(self, obj):
        """Validate the time window and usage."""
        now = tz_now()
        if obj.not_before_time and obj.not_before_time > now:
            return False
        if obj.expiration_time and obj.expiration_time < now:
            return False
        if obj.used_to_date >= obj.max_uses:
            return False
        return True
    is_valid.boolean = True


admin.site.register(RequestToken, RequestTokenAdmin)



# -*- coding: utf-8 -*-
"""request_token decorator tests."""
from datetime import datetime, timedelta

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.urlresolvers import reverse
from django.test import TransactionTestCase, Client

from request_token.models import RequestToken, RequestTokenLog
from request_token.settings import JWT_QUERYSTRING_ARG, JWT_SESSION_TOKEN_EXPIRY


def get_url(url_name, token):
    """Helper to format urls with tokens."""
    url = reverse('testing:%s' % url_name)
    if token:
        url += '?%s=%s' % (JWT_QUERYSTRING_ARG, token.jwt())
    return url


class ViewTests(TransactionTestCase):

    """Test the end-to-end use of tokens.

    These tests specifically confirm the way in which request and session tokens
    deal with repeated requests.

    """

    def setUp(self):
        self.client = Client()
        self.user = get_user_model().objects.create_user('zoidberg')

    def test_request_token(self):
        """Test the request tokens only set the user for a single request."""
        token = RequestToken.objects.create_token(
            scope="foo",
            max_uses=2,
            user=self.user,
            login_mode=RequestToken.LOGIN_MODE_REQUEST
        )
        response = self.client.get(get_url('decorated', token))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.request_user, self.user)
        self.assertEqual(RequestTokenLog.objects.count(), 1)

        response = self.client.get(get_url('undecorated', None))
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.request_user, AnonymousUser)
        self.assertEqual(RequestTokenLog.objects.count(), 1)

    def test_session_token(self):
        """Test that session tokens set the user for all requests."""
        token = RequestToken.objects.create_token(
            scope="foo",
            max_uses=1,
            user=self.user,
            login_mode=RequestToken.LOGIN_MODE_SESSION,
            expiration_time=(datetime.now() + timedelta(minutes=JWT_SESSION_TOKEN_EXPIRY))
        )

        response = self.client.get(get_url('decorated', token))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.request_user, self.user)
        self.assertEqual(RequestTokenLog.objects.count(), 1)

        # for a session token, all future requests should also be authenticated
        response = self.client.get(get_url('undecorated', None))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.request_user, self.user)
        self.assertEqual(RequestTokenLog.objects.count(), 1)



# -*- coding: utf-8 -*-
"""request_token decorator tests."""
from django.contrib.auth.models import AnonymousUser
from django.http import HttpResponse, HttpResponseForbidden
from django.test import TransactionTestCase, RequestFactory

from request_token.decorators import use_request_token, respond_to_error
from request_token.exceptions import ScopeError, TokenNotFoundError
from request_token.models import RequestToken, RequestTokenLog
from request_token.settings import JWT_QUERYSTRING_ARG
from request_token.middleware import RequestTokenMiddleware


@use_request_token(scope="foo")
def test_view_func(request):
    """Return decorated request / response objects."""
    response = HttpResponse("Hello, world!", status=200)
    return response


class MockSession(object):

    """Fake Session model used to support `session_key` property."""

    @property
    def session_key(self):
        return "foobar"


class DecoratorTests(TransactionTestCase):

    """use_jwt decorator tests."""

    def setUp(self):
        self.factory = RequestFactory()
        self.middleware = RequestTokenMiddleware()

    def _request(self, path, token, user):
        path = path + "?%s=%s" % (JWT_QUERYSTRING_ARG, token) if token else path
        request = self.factory.get(path)
        request.session = MockSession()
        request.user = user
        self.middleware.process_request(request)
        return request

    def test_respond_to_error(self):
        ex = Exception("foo")
        response = respond_to_error("bar", ex)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.error, ex)

    def test_no_token(self):
        request = self._request('/', None, AnonymousUser())
        response = test_view_func(request)
        self.assertEqual(response.status_code, 200)
        self.assertFalse(hasattr(request, 'token'))
        self.assertFalse(RequestTokenLog.objects.exists())

        # now force a TokenNotFoundError, by requiring it in the decorator
        @use_request_token(scope="foo", required=True)
        def test_view_func2(request):
            pass

        response = test_view_func2(request)
        self.assertIsInstance(response, HttpResponseForbidden)
        self.assertIsInstance(response.error, TokenNotFoundError)

    def test_scope(self):
        token = RequestToken.objects.create_token(scope="foobar")
        request = self._request('/', token.jwt(), AnonymousUser())
        response = test_view_func(request)
        self.assertIsInstance(response, HttpResponseForbidden)
        self.assertIsInstance(response.error, ScopeError)
        self.assertFalse(RequestTokenLog.objects.exists())

        RequestToken.objects.all().update(scope="foo")
        request = self._request('/', token.jwt(), AnonymousUser())
        response = test_view_func(request)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(RequestTokenLog.objects.exists())




# -*- coding: utf-8 -*-
"""request_token model tests."""
import datetime
import mock

from jwt.exceptions import InvalidAudienceError

from django.contrib.sessions.middleware import SessionMiddleware
from django.contrib.auth import get_user_model, logout
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.http import HttpResponse
from django.test import TransactionTestCase, RequestFactory
from django.utils.timezone import now as tz_now

from request_token.models import RequestToken, RequestTokenLog
from request_token.exceptions import MaxUseError
from request_token.settings import JWT_SESSION_TOKEN_EXPIRY
from request_token.utils import to_seconds, decode


class RequestTokenTests(TransactionTestCase):

    """RequestToken model property and method tests."""

    def setUp(self):
        self.user = get_user_model().objects.create_user('zoidberg')

    def test_defaults(self):
        token = RequestToken()
        self.assertIsNone(token.user)
        self.assertEqual(token.scope, '')
        self.assertEqual(token.login_mode, RequestToken.LOGIN_MODE_NONE)
        self.assertIsNone(token.expiration_time)
        self.assertIsNone(token.not_before_time)
        self.assertEqual(token.data, "{}")
        self.assertIsNone(token.issued_at)
        self.assertEqual(token.max_uses, 1)
        self.assertEqual(token.used_to_date, 0)

    def test_save(self):
        token = RequestToken().save()
        self.assertIsNotNone(token)
        self.assertIsNone(token.user)
        self.assertEqual(token.scope, '')
        self.assertEqual(token.login_mode, RequestToken.LOGIN_MODE_NONE)
        self.assertIsNone(token.expiration_time)
        self.assertIsNone(token.not_before_time)
        self.assertEqual(token.data, "{}")
        self.assertIsNotNone(token.issued_at)
        self.assertEqual(token.max_uses, 1)
        self.assertEqual(token.used_to_date, 0)

        token.issued_at = None
        token = token.save(update_fields=['issued_at'])
        self.assertIsNone(token.issued_at)

        now = tz_now()
        expires = now + datetime.timedelta(minutes=JWT_SESSION_TOKEN_EXPIRY)
        with mock.patch('request_token.models.tz_now', lambda: now):
            token = RequestToken(
                login_mode=RequestToken.LOGIN_MODE_SESSION,
                user=self.user,
                scope="foo"
            )
            self.assertIsNone(token.issued_at)
            self.assertIsNone(token.expiration_time)
            token.save()
            self.assertEqual(token.issued_at, now)
            self.assertEqual(token.expiration_time, expires)

    def test_claims(self):
        token = RequestToken()
        # raises error with no id set - put into context manager as it's
        # an attr, not a callable
        self.assertEqual(len(token.claims), 3)
        self.assertEqual(token.max, 1)
        self.assertEqual(token.sub, '')
        self.assertIsNone(token.jti)
        self.assertIsNone(token.aud)
        self.assertIsNone(token.exp)
        self.assertIsNone(token.nbf)
        self.assertIsNone(token.iat)

        # now let's set some properties
        token.user = self.user
        self.assertEqual(token.aud, self.user.id)
        self.assertEqual(len(token.claims), 4)

        token.login_mode = RequestToken.LOGIN_MODE_REQUEST
        self.assertEqual(token.claims['mod'], RequestToken.LOGIN_MODE_REQUEST[:1].lower())
        self.assertEqual(len(token.claims), 4)

        now = tz_now()
        now_sec = to_seconds(now)

        token.expiration_time = now
        self.assertEqual(token.exp, now_sec)
        self.assertEqual(len(token.claims), 5)

        token.not_before_time = now
        self.assertEqual(token.nbf, now_sec)
        self.assertEqual(len(token.claims), 6)

        # saving updates the id and issued_at timestamp
        with mock.patch('request_token.models.tz_now', lambda: now):
            token.save()
            self.assertEqual(token.iat, now_sec)
            self.assertEqual(token.jti, token.id)
            self.assertEqual(len(token.claims), 8)

    def test_clean(self):
        token = RequestToken(
            login_mode=RequestToken.LOGIN_MODE_NONE,
            # user=self.user
        )
        token.clean()
        # set a user, should now fail validation
        token.user = self.user
        self.assertRaises(ValidationError, token.clean)

        # request mode
        token.login_mode = RequestToken.LOGIN_MODE_REQUEST
        token.clean()
        token.user = None
        self.assertRaises(ValidationError, token.clean)

        def reset_session():
            """Reset properties so that token passes validation."""
            token.login_mode = RequestToken.LOGIN_MODE_SESSION
            token.user = self.user
            token.issued_at = tz_now()
            token.expiration_time = token.issued_at + datetime.timedelta(minutes=1)
            token.max_uses = 1

        def assertValidationFails(field_name):
            with self.assertRaises(ValidationError) as ctx:
                token.clean()
            self.assertTrue(field_name in dict(ctx.exception))

        # check the rest_session works!
        reset_session()
        token.clean()
        token.max_uses = 10
        assertValidationFails('max_uses')

        reset_session()
        token.user = None
        assertValidationFails('user')

        reset_session()
        token.expiration_time = None
        assertValidationFails('expiration_time')

    def test_log(self):
        token = RequestToken().save()
        factory = RequestFactory()
        request = factory.get('/')
        request.user = AnonymousUser()
        request.META = {}
        response = HttpResponse("foo", status=123)

        def assertUsedToDate(expected):
            token.refresh_from_db(fields=['used_to_date'])
            self.assertEqual(token.used_to_date, expected)

        log = token.log(request, response)
        self.assertEqual(RequestTokenLog.objects.get(), log)
        self.assertEqual(log.user, None)
        self.assertEqual(log.token, token)
        self.assertEqual(log.user_agent, 'unknown')
        self.assertEqual(log.client_ip, 'unknown')
        self.assertEqual(log.status_code, 123)
        assertUsedToDate(1)

        request.META['REMOTE_ADDR'] = '192.168.0.1'
        log = token.log(request, response)
        self.assertEqual(log.client_ip, '192.168.0.1')
        assertUsedToDate(2)

        request.META['HTTP_X_FORWARDED_FOR'] = '192.168.0.2'
        log = token.log(request, response)
        self.assertEqual(log.client_ip, '192.168.0.2')
        assertUsedToDate(3)

        request.META['HTTP_USER_AGENT'] = 'test_agent'
        log = token.log(request, response)
        self.assertEqual(log.user_agent, 'test_agent')
        token.refresh_from_db(fields=['used_to_date'])
        assertUsedToDate(4)

    def test_jwt(self):
        token = RequestToken(id=1, scope='foo').save()
        jwt = token.jwt()
        self.assertEqual(decode(jwt), token.claims)

    def test_validate_max_uses(self):
        token = RequestToken(max_uses=1, used_to_date=0)
        token.validate_max_uses()
        token.used_to_date = token.max_uses
        self.assertRaises(MaxUseError, token.validate_max_uses)

    def test__auth_is_anonymous(self):
        factory = RequestFactory()
        middleware = SessionMiddleware()
        anon = AnonymousUser()
        request = factory.get('/foo')
        middleware.process_request(request)
        request.user = anon

        # try default token
        token = RequestToken.objects.create_token(
            scope="foo",
            max_uses=10,
            login_mode=RequestToken.LOGIN_MODE_NONE
        )
        request = token._auth_is_anonymous(request)
        self.assertEqual(request.user, anon)

        # try request token
        user1 = get_user_model().objects.create_user(username="Finbar")
        token = RequestToken.objects.create_token(
            user=user1,
            scope="foo",
            max_uses=10,
            login_mode=RequestToken.LOGIN_MODE_REQUEST
        )
        token._auth_is_anonymous(request)
        self.assertEqual(request.user, user1)
        self.assertFalse(hasattr(token.user, 'backend'))

        # try a session token
        logout(request)
        request.user = anon
        token.login_mode = RequestToken.LOGIN_MODE_SESSION
        request = token._auth_is_anonymous(request)
        self.assertEqual(request.user, user1)
        self.assertEqual(token.user.backend, 'django.contrib.auth.backends.ModelBackend')

    def test__auth_is_authenticated(self):
        factory = RequestFactory()
        middleware = SessionMiddleware()
        request = factory.get('/foo')
        middleware.process_request(request)
        user1 = get_user_model().objects.create_user(username="Jekyll")
        request.user = user1

        # try default token
        token = RequestToken.objects.create_token(
            scope="foo",
            max_uses=10,
            login_mode=RequestToken.LOGIN_MODE_NONE
        )
        request = token._auth_is_authenticated(request)
        self.assertEqual(request.user, user1)

        # try request token
        token = RequestToken.objects.create_token(
            user=user1,
            scope="foo",
            max_uses=10,
            login_mode=RequestToken.LOGIN_MODE_REQUEST
        )
        request = token._auth_is_authenticated(request)

        token.login_mode = RequestToken.LOGIN_MODE_SESSION
        request = token._auth_is_authenticated(request)
        self.assertEqual(request.user, user1)

        token.user = get_user_model().objects.create_user(username="Hyde")
        self.assertRaises(InvalidAudienceError, token._auth_is_authenticated, request)

    def test_authenticate(self):
        factory = RequestFactory()
        middleware = SessionMiddleware()
        anon = AnonymousUser()
        request = factory.get('/foo')
        middleware.process_request(request)
        request.user = anon

        user1 = get_user_model().objects.create_user(username="Finbar")
        token = RequestToken.objects.create_token(
            user=user1,
            scope="foo",
            max_uses=10,
            login_mode=RequestToken.LOGIN_MODE_REQUEST
        )
        token.authenticate(request)
        self.assertEqual(request.user, user1)

        request.user = get_user_model().objects.create_user(username="Hyde")
        self.assertRaises(InvalidAudienceError, token.authenticate, request)


class RequestTokenQuerySetTests(TransactionTestCase):

    """RequestTokenQuerySet class tests."""

    def test_create_token(self):
        self.assertRaises(TypeError, RequestToken.objects.create_token)
        RequestToken.objects.create_token(scope="foo")
        self.assertEqual(RequestToken.objects.get().scope, 'foo')


class RequestTokenLogTests(TransactionTestCase):

    """RequestTokenLog model property and method tests."""

    def setUp(self):
        self.user = get_user_model().objects.create_user('zoidberg')
        self.token = RequestToken.objects.create_token(
            scope='foo',
            user=self.user,
            login_mode=RequestToken.LOGIN_MODE_REQUEST
        )

    def test_defaults(self):
        log = RequestTokenLog(
            token=self.token,
            user=self.user
        )
        self.assertEqual(log.user, self.user)
        self.assertEqual(log.token, self.token)
        self.assertEqual(log.user_agent, '')
        self.assertEqual(log.client_ip, '')
        self.assertIsNone(log.timestamp)

    def test_save(self):
        log = RequestTokenLog(
            token=self.token,
            user=self.user
        ).save()
        self.assertIsNotNone(log.timestamp)

        log.timestamp = None
        self.assertRaises(IntegrityError, log.save, update_fields=['timestamp'])



# -*- coding:utf-8 -*-
from django.http import HttpResponse

from request_token.decorators import use_request_token


def undecorated(request):
    response = HttpResponse(u"Hello, %s" % request.user)
    response.request_user = request.user
    return response


@use_request_token(scope="foo")
def decorated(request):
    response = HttpResponse(u"Hello, %s" % request.user)
    response.request_user = request.user
    return response


# -*- coding: utf-8 -*-
"""request_token settings."""
from django.conf import settings

# the name of GET argument to contain the token
JWT_QUERYSTRING_ARG = getattr(settings, 'JWT_QUERYSTRING_ARG', 'rt')

# the fixed expiration check on Session tokens
JWT_SESSION_TOKEN_EXPIRY = int(getattr(settings, 'JWT_SESSION_TOKEN_EXPIRY', 10))

# Set the default 403 template value
FOUR03_TEMPLATE = getattr(settings, 'FOUR03_TEMPLATE', None)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='RequestToken',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('target_url', models.CharField(help_text=b'The target endpoint.', max_length=200)),
                ('expiration_time', models.DateTimeField(help_text=b'DateTime at which this token expires.', null=True, blank=True)),
                ('not_before_time', models.DateTimeField(help_text=b'DateTime before which this token is invalid.', null=True, blank=True)),
                ('data', models.TextField(help_text=b'Custom data (JSON) added to the default payload.', max_length=1000, blank=True)),
                ('issued_at', models.DateTimeField(help_text=b'Time the token was created, set in the initial save.', null=True, blank=True)),
                ('max_uses', models.IntegerField(default=1, help_text=b'Cap on the number of times the token can be used, defaults to 1 (single use).')),
                ('used_to_date', models.IntegerField(default=0, help_text=b'Denormalised count of the number times the token has been used.')),
                ('user', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, help_text=b'Intended recipient of the JWT.', null=True)),
            ],
        ),
        migrations.CreateModel(
            name='RequestTokenLog',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('user_agent', models.TextField(help_text=b'User-agent of client used to make the request.', blank=True)),
                ('client_ip', models.CharField(help_text=b'Client IP of device used to make the request.', max_length=15)),
                ('timestamp', models.DateTimeField(help_text=b'Time the request was logged.')),
                ('token', models.ForeignKey(help_text=b'The RequestToken that was used.', to='request_token.RequestToken')),
                ('user', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, help_text=b'The user who made the request (None if anonymous).', null=True)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('request_token', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='requesttokenlog',
            name='status_code',
            field=models.IntegerField(help_text=b'Response status code associated with this use of the token.', null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='requesttoken',
            name='data',
            field=models.TextField(default=b'{}', help_text=b'Custom data (JSON) added to the default payload.', max_length=1000, blank=True),
        ),
        migrations.AlterField(
            model_name='requesttokenlog',
            name='timestamp',
            field=models.DateTimeField(help_text=b'Time the request was logged.', blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
"""request_token models."""
import datetime
import logging

from django.contrib.auth import login
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.utils.timezone import now as tz_now

from jwt.exceptions import InvalidAudienceError

from request_token.exceptions import MaxUseError
from request_token.settings import JWT_SESSION_TOKEN_EXPIRY
from request_token.utils import to_seconds, encode

logger = logging.getLogger(__name__)


class RequestTokenQuerySet(models.query.QuerySet):

    """Custom QuerySet for RquestToken objects."""

    def create_token(self, scope, **kwargs):
        """Create a new RequestToken."""
        return RequestToken(scope=scope, **kwargs).save()


class RequestToken(models.Model):

    """A link token, targeted for use by a known Django User.

    A RequestToken contains information that can be encoded as a JWT
    (JSON Web Token). It is designed to be used in conjunction with the
    RequestTokenMiddleware (responsible for JWT verification) and the
    @use_request_token decorator (responsible for validating the token
    and setting the request.user correctly).

    Each token must have a 'scope', which is used to tie it to a view function
    that is decorated with the `use_request_token` decorator. The token can
    only be used by functions with matching scopes.

    The token may be set to a specific User, in which case, if the existing
    request is unauthenticated, it will use that user as the `request.user`
    property, allowing access to authenticated views.

    The token may be timebound by the `not_before_time` and `expiration_time`
    properties, which are registered JWT 'claims'.

    The token may be restricted by the number of times it can be used, through
    the `max_use` property, which is incremented each time it's used (NB *not*
    thread-safe).

    The token may also store arbitrary serializable data, which can be used
    by the view function if the request token is valid.

    JWT spec: https://tools.ietf.org/html/rfc7519

    """

    # do not login the user on the request
    LOGIN_MODE_NONE = 'None'
    # login the user, but only for the original request
    LOGIN_MODE_REQUEST = 'Request'
    # login the user fully, but only for single-use short-duration links
    LOGIN_MODE_SESSION = 'Session'

    LOGIN_MODE_CHOICES = (
        (LOGIN_MODE_NONE, 'Do not authenticate'),
        (LOGIN_MODE_REQUEST, 'Authenticate a single request'),
        (LOGIN_MODE_SESSION, 'Authenticate for the entire session'),
    )
    login_mode = models.CharField(
        max_length=10,
        default=LOGIN_MODE_NONE,
        choices=LOGIN_MODE_CHOICES,
        help_text="How should the request be authenticated?"
    )
    user = models.ForeignKey(
        User,
        blank=True, null=True,
        help_text="Intended recipient of the JWT (can be used by anyone if not set)."
    )
    scope = models.CharField(
        max_length=100,
        help_text="Label used to match request to view function in decorator."
    )
    expiration_time = models.DateTimeField(
        blank=True, null=True,
        help_text="Token will expire at this time (raises ExpiredSignatureError)."
    )
    not_before_time = models.DateTimeField(
        blank=True, null=True,
        help_text="Token cannot be used before this time (raises ImmatureSignatureError)."
    )
    data = models.TextField(
        max_length=1000,
        help_text="Custom data add to the token, but not encoded (must be fetched from DB).",
        blank=True,
        default='{}'
    )
    issued_at = models.DateTimeField(
        blank=True, null=True,
        help_text="Time the token was created (set in the initial save)."
    )
    max_uses = models.IntegerField(
        default=1,
        help_text="The maximum number of times the token can be used."
    )
    used_to_date = models.IntegerField(
        default=0,
        help_text="Number of times the token has been used to date (raises MaxUseError)."
    )

    objects = RequestTokenQuerySet.as_manager()

    def __unicode__(self):
        return u"Request token #%s" % (self.id)

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __repr__(self):
        return u"<RequestToken id=%s scope=%s login_mode='%s'>" %(
            self.id, self.scope, self.login_mode)

    @property
    def aud(self):
        """The 'aud' claim, maps to user.id."""
        return self.claims.get('aud')

    @property
    def exp(self):
        """The 'exp' claim, maps to expiration_time."""
        return self.claims.get('exp')

    @property
    def nbf(self):
        """The 'nbf' claim, maps to not_before_time."""
        return self.claims.get('nbf')

    @property
    def iat(self):
        """The 'iat' claim, maps to issued_at."""
        return self.claims.get('iat')

    @property
    def jti(self):
        """The 'jti' claim, maps to id."""
        return self.claims.get('jti')

    @property
    def max(self):
        """The 'max' claim, maps to max_uses."""
        return self.claims.get('max')

    @property
    def sub(self):
        """The 'sub' claim, maps to scope."""
        return self.claims.get('sub')

    @property
    def claims(self):
        """A dict containing all of the DEFAULT_CLAIMS (where values exist)."""
        claims = {
            'max': self.max_uses,
            'sub': self.scope,
            'mod': self.login_mode[:1].lower()
        }
        if self.id is not None:
            claims['jti'] = self.id
        if self.user is not None:
            claims['aud'] = self.user.id
        if self.expiration_time is not None:
            claims['exp'] = to_seconds(self.expiration_time)
        if self.issued_at is not None:
            claims['iat'] = to_seconds(self.issued_at)
        if self.not_before_time is not None:
            claims['nbf'] = to_seconds(self.not_before_time)
        return claims

    def clean(self):
        """Ensure that login_mode setting is valid."""
        if self.login_mode == RequestToken.LOGIN_MODE_NONE:
            if self.user is not None:
                raise ValidationError(
                    {'user': u"User must be None if login_mode is LOGIN_MODE_NONE."}
                )
        if self.login_mode == RequestToken.LOGIN_MODE_SESSION:
            if self.user is None:
                raise ValidationError(
                    {'user': u"Session token must have a user."}
                )
            if self.max_uses != 1:
                raise ValidationError(
                    {'max_uses': u"Session token must have max_use of 1."}
                )
            if self.expiration_time is None:
                raise ValidationError(
                    {'expiration_time': u"Session token must have an expiration_time."}
                )
        if self.login_mode == RequestToken.LOGIN_MODE_REQUEST:
            if self.user is None:
                raise ValidationError(
                    {'expiration_time': u"Request token must have a user."}
                )

    def save(self, *args, **kwargs):
        if 'update_fields' not in kwargs:
            self.issued_at = self.issued_at or tz_now()
            if self.login_mode == RequestToken.LOGIN_MODE_SESSION:
                self.expiration_time = self.expiration_time or (
                    self.issued_at +
                    datetime.timedelta(minutes=JWT_SESSION_TOKEN_EXPIRY)
                )
        self.clean()
        super(RequestToken, self).save(*args, **kwargs)
        return self

    def jwt(self):
        """Encode the token claims into a JWT."""
        return encode(self.claims)

    def validate_max_uses(self):
        """Check the token max_uses is still valid.

        Raises MaxUseError if invalid.

        """
        if self.used_to_date >= self.max_uses:
            raise MaxUseError(
                u"RequestToken [%s] has exceeded max uses" % self.id
            )

    def _auth_is_anonymous(self, request):
        """Authenticate anonymous requests."""
        assert request.user.is_anonymous(), u"User is authenticated."

        if self.login_mode == RequestToken.LOGIN_MODE_NONE:
            pass

        if self.login_mode == RequestToken.LOGIN_MODE_REQUEST:
            logger.debug(
                u"Setting request.user to %r from token %i.",
                self.user, self.id
            )
            request.user = self.user

        if self.login_mode == RequestToken.LOGIN_MODE_SESSION:
            logger.debug(
                u"Authenticating request.user as %r from token %i.",
                self.user, self.id
            )
            # I _think_ we can get away with this as we are pulling the
            # user out of the DB, and we are explicitly authenticating
            # the user.
            self.user.backend = 'django.contrib.auth.backends.ModelBackend'
            login(request, self.user)

        return request

    def _auth_is_authenticated(self, request):
        """Authenticate requests with existing users."""
        assert request.user.is_authenticated(), u"User is anonymous."

        if self.login_mode == RequestToken.LOGIN_MODE_NONE:
            return request

        if request.user == self.user:
            return request

        raise InvalidAudienceError(
            "RequestToken [%i] audience mismatch: '%s' != '%s'" %
            (self.id, request.user, self.user)
        )

    def authenticate(self, request):
        """Authenticate an HttpRequest with the token user.

        This method encapsulates the request handling - if the token
        has a user assigned, then this will be added to the request.

        """
        if request.user.is_anonymous():
            return self._auth_is_anonymous(request)
        else:
            return self._auth_is_authenticated(request)

    @transaction.atomic
    def log(self, request, response):
        """Record the use of a token.

        This is used by the decorator to log each time someone uses the token,
        or tries to. Used for reporting, diagnostics.

        Args:
            request: the HttpRequest object that used the token, from which the
                user, ip and user-agenct are extracted.
            response: the corresponding HttpResponse object, from which the status
                code is extracted.
            duration: float, the duration of the view function in ms - just because
                you can never measure too many things.

        Returns a RequestTokenUse object.

        """
        assert hasattr(request, 'user'), (
            "Request is missing user property. Please ensure that the Django "
            "authentication middleware is installed."
        )
        meta = request.META
        xff = meta.get('HTTP_X_FORWARDED_FOR', None)
        client_ip = xff or meta.get('REMOTE_ADDR', 'unknown')
        user = None if request.user.is_anonymous() else request.user
        rtu = RequestTokenLog(
            token=self,
            user=user,
            user_agent=meta.get('HTTP_USER_AGENT', 'unknown'),
            client_ip=client_ip,
            status_code=response.status_code
        ).save()
        # NB this could already be out-of-date
        self.used_to_date = models.F('used_to_date') + 1
        self.save()
        return rtu


class RequestTokenLog(models.Model):

    """Used to log the use of a RequestToken."""

    token = models.ForeignKey(
        RequestToken,
        help_text="The RequestToken that was used.",
        db_index=True
    )
    user = models.ForeignKey(
        User,
        blank=True, null=True,
        help_text="The user who made the request (None if anonymous)."
    )
    user_agent = models.TextField(
        blank=True,
        help_text="User-agent of client used to make the request."
    )
    client_ip = models.CharField(
        max_length=15,
        help_text="Client IP of device used to make the request."
    )
    status_code = models.IntegerField(
        blank=True, null=True,
        help_text="Response status code associated with this use of the token."
    )
    timestamp = models.DateTimeField(
        blank=True,
        help_text="Time the request was logged."
    )

    def __unicode__(self):
        if self.user is None:
            return u"%s used %s" % (self.token, self.timestamp)
        else:
            return u"%s used by %s at %s" % (self.token, self.user, self.timestamp)

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __repr__(self):
        return u"<RequestTokenLog id=%s token=%s timestamp='%s'>" %(
            self.id, self.token.id, self.timestamp)

    def save(self, *args, **kwargs):
        if 'update_fields' not in kwargs:
            self.timestamp = self.timestamp or tz_now()
        super(RequestTokenLog, self).save(*args, **kwargs)
        return self



# -*- coding: utf8 -*-
"""request_token expections.

Local exceptions related to tokens inherit from the PyJWT base
InvalidTokenError.

"""
from jwt.exceptions import InvalidTokenError


class MaxUseError(InvalidTokenError):

    """Error raised when a token has exceeded its max_use cap."""

    pass


class ScopeError(InvalidTokenError):

    """Error raised when a token scope does not match the function scope."""

    pass


class TokenNotFoundError(InvalidTokenError):

    """Error raised when a token is expected, but not found."""

    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('request_token', '0003_auto_20151229_1105'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='requesttoken',
            name='target_url',
        ),
    ]



# -*- coding: utf-8 -*-
"""request_token.utils tests."""
import datetime

from jwt import decode as jwt_decode, encode as jwt_encode
from jwt.exceptions import MissingRequiredClaimError, DecodeError

from django.conf import settings
from django.test import TestCase

from request_token.utils import encode, decode, to_seconds, MANDATORY_CLAIMS


class FunctionTests(TestCase):

    """Tests for free-floating functions."""

    def test_to_seconds(self):
        timestamp = datetime.datetime(2015, 1, 1)
        self.assertEqual(to_seconds(timestamp), 1420070400)
        self.assertEqual(to_seconds(1420070400), None)

    def test_encode(self):
        payload = {'foo': 'bar'}
        self.assertRaises(MissingRequiredClaimError, encode, payload)
        # force all mandatory claims into the payload
        payload = {k: 'foo' for k in MANDATORY_CLAIMS}
        self.assertEqual(encode(payload), jwt_encode(payload, settings.SECRET_KEY))

    def test_decode(self):
        # check that we can't decode with the wrong secret
        payload = {'foo': 'bar'}
        encoded = jwt_encode(payload, "QWERTYUIO")
        self.assertRaises(DecodeError, decode, encoded)
        # we can decode this, but we're missing the mandatory fields
        encoded = jwt_encode(payload, settings.SECRET_KEY)
        self.assertRaises(MissingRequiredClaimError, decode, encoded)
        # force all mandatory claims into the payload
        payload = {k: 'foo' for k in MANDATORY_CLAIMS}
        encoded = jwt_encode(payload, settings.SECRET_KEY)
        self.assertEqual(decode(encoded), payload)
        self.assertEqual(decode(encoded), jwt_decode(encoded, settings.SECRET_KEY))



# -*- coding: utf-8 -*-
"""request_token middleware tests."""
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.http import HttpResponseForbidden, HttpResponseNotAllowed
from django.test import TransactionTestCase, RequestFactory

from jwt import exceptions

from request_token.models import RequestToken
from request_token.settings import JWT_QUERYSTRING_ARG
from request_token.middleware import RequestTokenMiddleware


class MockSession(object):

    """Fake Session model used to support `session_key` property."""

    @property
    def session_key(self):
        return "foobar"


class MiddlewareTests(TransactionTestCase):

    """RequestTokenMiddleware tests."""

    def setUp(self):
        self.user = get_user_model().objects.create_user('zoidberg')
        self.factory = RequestFactory()
        self.middleware = RequestTokenMiddleware()

    def test_process_request_assertions(self):
        request = self.factory.get('/')
        middleware = self.middleware

        process_request = middleware.process_request
        self.assertRaises(AssertionError, process_request, request)

        request.user = AnonymousUser()
        self.assertRaises(AssertionError, process_request, request)
        request.session = MockSession()

        self.assertIsNone(process_request(request))
        self.assertFalse(hasattr(request, 'token_payload'))

    def test_process_request_without_token(self):
        request = self.factory.post('/')
        process_request = self.middleware.process_request
        request = self.factory.get('/')
        request.user = AnonymousUser()
        request.session = MockSession()
        self.assertIsNone(process_request(request))
        self.assertFalse(hasattr(request, 'token_payload'))

    def test_process_request_with_token(self):

        def new_request(jwt, user=self.user, session=MockSession()):
            request = self.factory.get('/?%s=%s' % (JWT_QUERYSTRING_ARG, jwt))
            request.user = user
            request.session = session
            return request

        # has a valid token
        token = RequestToken.objects.create_token(scope="foo")
        request = new_request(token.jwt())
        self.assertIsNone(self.middleware.process_request(request))
        self.assertEqual(request.token, token)

    def test_process_request_not_allowed(self):

        def new_request(jwt, user=self.user, session=MockSession()):
            request = self.factory.post('/?%s=%s' % (JWT_QUERYSTRING_ARG, jwt))
            request.user = user
            request.session = session
            return request

        # has a valid token
        token = RequestToken(id=1, scope="foo")
        request = new_request(token.jwt())
        response = self.middleware.process_request(request)
        self.assertIsInstance(response, HttpResponseNotAllowed)
        self.assertFalse(hasattr(response, 'error'))
        self.assertFalse(hasattr(request, 'token_payload'))
        self.assertEqual(response.status_code, 405)

    def test_process_request_forbidden(self):

        def new_request(jwt, user=self.user, session=MockSession()):
            request = self.factory.get('/?%s=%s' % (JWT_QUERYSTRING_ARG, jwt))
            request.user = user
            request.session = session
            return request

        # has an invalid token
        request = new_request("foo")  # this won't decode
        response = self.middleware.process_request(request)
        self.assertIsInstance(response, HttpResponseForbidden)
        self.assertIsInstance(response.error, exceptions.DecodeError)
        self.assertEqual(response.status_code, 403)
        self.assertFalse(hasattr(request, 'token_payload'))


# class IntegrationTests(TransactionTestCase):

#     """Test the end-to-end process."""




# -*- coding: utf-8 -*-
from django.conf.urls import patterns, url

urlpatterns = patterns(
    'test_app.views',
    url(r'^decorated/$', 'decorated', name="decorated"),
    url(r'^undecorated/$', 'undecorated', name="undecorated"),
)



# -*- coding: utf-8 -*-
"""Basic encode/decode utils, taken from PyJWT."""
import calendar

from jwt import encode as jwt_encode
from jwt import decode as jwt_decode
from jwt import exceptions

from django.conf import settings

# verification options - signature and expiry date
DEFAULT_DECODE_OPTIONS = {
    'verify_signature': True,
    'verify_exp': True,
    'verify_nbf': True,
    'verify_iat': True,
    'verify_aud': False,
    'verify_iss': False,  # we're only validating our own claims
    'require_exp': False,
    'require_iat': False,
    'require_nbf': False
}

MANDATORY_CLAIMS = ('jti', 'sub', 'mod')


def check_mandatory_claims(payload, claims=MANDATORY_CLAIMS):
    """Check dict for mandatory claims."""
    for claim in claims:
        if claim not in payload:
            raise exceptions.MissingRequiredClaimError(claim)


def encode(payload, check_claims=MANDATORY_CLAIMS):
    """Encode JSON payload (using SECRET_KEY)."""
    check_mandatory_claims(payload, claims=check_claims)
    return jwt_encode(payload, settings.SECRET_KEY)


def decode(token, options=DEFAULT_DECODE_OPTIONS, check_claims=MANDATORY_CLAIMS):
    """Decode JWT payload and check for 'jti', 'sub' claims."""
    decoded = jwt_decode(token, settings.SECRET_KEY, options=options)
    check_mandatory_claims(decoded, claims=check_claims)
    return decoded


def to_seconds(timestamp):
    """Convert timestamp into integers since epoch."""
    try:
        return calendar.timegm(timestamp.utctimetuple())
    except:
        return None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('request_token', '0004_remove_requesttoken_target_url'),
    ]

    operations = [
        migrations.AddField(
            model_name='requesttoken',
            name='login_mode',
            field=models.CharField(default=b'None', help_text=b'How should the request be authenticated?', max_length=10, choices=[(b'None', b'Do not authenticate'), (b'Request', b'Authenticate a single request'), (b'Session', b'Authenticate for the entire session')]),
        ),
        migrations.AlterField(
            model_name='requesttoken',
            name='data',
            field=models.TextField(default=b'{}', help_text=b'Custom data add to the token, but not encoded (must be fetched from DB).', max_length=1000, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('request_token', '0002_auto_20151227_1428'),
    ]

    operations = [
        migrations.AddField(
            model_name='requesttoken',
            name='scope',
            field=models.CharField(default='', help_text=b'Label used to match request to view function in decorator.', max_length=100),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='requesttoken',
            name='expiration_time',
            field=models.DateTimeField(help_text=b'Token will expire at this time (raises ExpiredSignatureError).', null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='requesttoken',
            name='issued_at',
            field=models.DateTimeField(help_text=b'Time the token was created (set in the initial save).', null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='requesttoken',
            name='max_uses',
            field=models.IntegerField(default=1, help_text=b'The maximum number of times the token can be used.'),
        ),
        migrations.AlterField(
            model_name='requesttoken',
            name='not_before_time',
            field=models.DateTimeField(help_text=b'Token cannot be used before this time (raises ImmatureSignatureError).', null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='requesttoken',
            name='used_to_date',
            field=models.IntegerField(default=0, help_text=b'Number of times the token has been used to date (raises MaxUseError).'),
        ),
        migrations.AlterField(
            model_name='requesttoken',
            name='user',
            field=models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, help_text=b'Intended recipient of the JWT (can be used by anyone if not set).', null=True),
        ),
    ]



# -*- coding: utf-8 -*-
"""request_token middleware."""
import logging

from django.http import HttpResponseForbidden, HttpResponseNotAllowed

from jwt.exceptions import InvalidTokenError

from request_token.models import RequestToken
from request_token.settings import JWT_QUERYSTRING_ARG
from request_token.utils import decode

## For rendering  a custom 403-page
from django.template import loader 
from request_token.settings import FOUR03_TEMPLATE

logger = logging.getLogger(__name__)


class RequestTokenMiddleware(object):

    """Extract and verify request tokens from incoming GET requests.

    This middleware is used to perform initial JWT verfication of
    link tokens.

    """

    def process_request(self, request):
        """Verify JWT request querystring arg.

        If a token is found (using JWT_QUERYSTRING_ARG), then it is decoded,
        which verifies the signature and expiry dates, and raises a 403 if
        the token is invalid.

        The decoded payload is then added to the request as the `token_payload`
        property - allowing it to be interrogated by the view function
        decorator when it gets there.

        We don't substitute in the user at this point, as we are not making
        any assumptions about the request path at this point - it's not until
        we get to the view function that we know where we are heading - at
        which point we verify that the scope matches, and only then do we
        use the token user.

        """
        assert hasattr(request, 'session'), (
            "Request has no session attribute, please ensure that Django "
            "session middleware is installed."
        )
        assert hasattr(request, 'user'), (
            "Request has no user attribute, please ensure that Django "
            "authentication middleware is installed."
        )

        token = request.GET.get(JWT_QUERYSTRING_ARG)

        if token is None:
            return

        if request.method != 'GET':
            return HttpResponseNotAllowed(['GET'])

        try:
            payload = decode(token)
            token = RequestToken.objects.get(id=payload['jti'])
            token.validate_max_uses()
            token.authenticate(request)
            request.token = token

        except (RequestToken.DoesNotExist, InvalidTokenError) as ex:
            key = request.session.session_key
            logger.warning(
                "JWT token error (error code:'%s'): %s", key, ex
            )
            if FOUR03_TEMPLATE:
                response = HttpResponseForbidden(
                    loader.render_to_string(FOUR03_TEMPLATE, context = {'token_error':'Temporary link token error: %s'%key})
                )
            else:
                response = HttpResponseForbidden(
                    u"Temporary link token error (code: %s)" % key
                )
            response.error = ex
            return response
